import statsmodels.api as sm
import numpy as np


def fit_polynomial(order, X, Y):
    """
        Fits a polinomia to data, with numpy

        Arguments:
            order: int. Order of the polinomial
            X, Y: 1d np.ndarray with data

        Reurns:
            f: np.array with fitted polynomial
    """
    #  calculate polynomial
    z = np.polyfit(X, Y, order)
    f = np.poly1d(z)
    return f


def linear_regression(X, Y, robust=False, rais_on_nan=False):
    """
        Fits a linear regression and returns the results and parameters

        Arguments:
            X, Y: np.ndarrays with data
            robust: bool. If trye robust RLM algorithm is fitted
            rais_on_nan: bool. If true an error is raised if there's nans
                in the input data

        Returns:
            X: np.array. With nans removed
            slope: float. SLope of linear regression
            intercept: float. Intercept of linear regressoin
            res: Results object with fitted model
    """
    # remove NANs
    remove_idx = [
        i for i, (x, y) in enumerate(zip(X, Y)) if np.isnan(x) or np.isnan(y)
    ]
    if rais_on_nan and remove_idx:
        raise ValueError(
            "Found invalid values in input data during linear regression"
        )

    X = np.delete(X, remove_idx)
    Y = np.delete(Y, remove_idx)

    X = sm.add_constant(X)
    if robust:
        res = sm.RLM(Y, X, missing="drop").fit()
    else:
        res = sm.OLS(Y, X).fit()  # Ordinary least squares

    return X, res.params[0], res.params[1], res


def fit_kde(X, **kwargs):
    """ 
        Fit a KDE using StatsModels. 

        Arguments:
            X: 1d np array with data to fit the KDE to.
            kwargs is useful to pass stuff to the fit, e.g. the binwidth (bw)

    """
    X = np.array(X).astype(np.float)
    kde = sm.nonparametric.KDEUnivariate(X)
    kde.fit(**kwargs)  # Estimate the densities

    return kde



import numpy as np
from scipy import stats
from collections import namedtuple
import pandas as pd
from scipy.signal import resample


def resample_list_of_arrayes_to_avg_len(lst, N=None, interpolate=False):
    """
        Given a list of arrays of varying length, this function
        resamples them so that they all have the 
        average length.
        Then it returns the vstack of the array
    """
    if N is None:
        N = np.mean([len(x) for x in lst]).astype(np.int32)

    if interpolate:
        lst = [pd.Series(x).interpolate() for x in lst]

    return np.vstack([resample(X, N) for X in lst])


def unwrap(X, isdeg=True):
    """
        Unwraps an array of angular data.

        Arguments:
            X: np.array with data
            isdeg: bool. If true the data is in degrees
    """
    if isdeg:
        X = np.radians(X)

    X = np.unwrap(X)

    if isdeg:
        X = np.degrees(X)

    return X


def find_nearest(X, value):
    """
        Finds the index of the array value that is closest to a target value.

        Arguments:
            X: np.ndarray with data
            value: float, int. Value
    """

    idx = np.abs(X - value).argmin()
    return idx


def percentile_range(X, low=5, high=95):
    """ 
        Calculates the range between the low and high percentiles
        in an array of data X
        
        Arguments:
            X: data
            low, high: int. Low and high percentiles
    """

    lowp = np.nanpercentile(X, low)
    highp = np.nanpercentile(X, high)
    median = np.nanmedian(X)
    mean = np.nanmean(X)
    std = np.std(X)
    sem = stats.sem(X)

    res = namedtuple("percentile", "low median mean high std sem")
    return res(lowp, median, mean, highp, std, sem)



import cv2
from loguru import logger

from fcutils._video import open_video
from fcutils._file import pathify, raise_on_path_not_exists

# ------------------------ Create opencv video writers ----------------------- #


def open_cvwriter(
    filepath, w=None, h=None, framerate=None, format="mp4", iscolor=False
):
    """
        Creats an instance of cv.VideoWriter to write video to file

        Arguments:
            filepath: str, Path. Where video is to be saved
            w,h: int. width and height of frame in pixels
            framerate: int. fps of output video
            format: video format ('mp4' or 'avi')
            iscolor: bool, set as true if images are rgb, else false if they are gray

        Returns:
            cv2.VideoWriter
    """
    try:
        if "avi" in format:
            fourcc = cv2.VideoWriter_fourcc("M", "J", "P", "G")
        else:
            fourcc = cv2.VideoWriter_fourcc(*"mp4v")
        videowriter = cv2.VideoWriter(
            filepath, fourcc, framerate, (w, h), iscolor
        )
    except:
        raise ValueError("Could not create videowriter")
    else:
        return videowriter


# --------------------- Manipulate opened video captures --------------------- #
def cap_set_frame(cap, frame_number):
    """
        Sets an opencv video capture object to a specific frame

        Arguments:
            cap: a VideoCapture object
            frame_number: int frame number
    """
    cap.set(1, frame_number)


def get_cap_selected_frame(cap, frame_number):
    """ 
        Gets a frame from an opencv video capture object to a specific frame

        Arguments:
            cap: a VideoCapture object
            frame_number: int frame number
    """
    cap_set_frame(cap, frame_number)
    ret, frame = cap.read()

    if not ret:
        logger.debug(f"FCUTILS: failed to read frame {frame_number}.")
        return None
    else:
        return frame


@open_video
def get_video_params(video):
    """ 
        Gets video parameters from an opencv video capture object
        
        Arguments:
            video: can be either the path to a video or a videocapture object

        Returns:
            nframes: int. Number of frames in video
            width, height: int. Frame size
            fps: int. Framerate
            is_color: bool. True if video si RGB false if it's 1 color.
    """
    frame = get_cap_selected_frame(video, 0)

    if frame is None:
        raise ValueError(
            "Could not read frame from cap while getting video params"
        )

    if frame.shape[1] == 3 or frame.shape[0] == 3:
        is_color = True
    else:
        is_color = False
    cap_set_frame(video, 0)

    nframes = int(video.get(cv2.CAP_PROP_FRAME_COUNT))
    width = int(video.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(video.get(cv2.CAP_PROP_FRAME_HEIGHT))
    fps = int(video.get(cv2.CAP_PROP_FPS))
    return nframes, width, height, fps, is_color


# ---------------------------- open videocaptures ---------------------------- #


def get_cap_from_file(path):
    '''
        Opens a cv2 video capture from a file path
    '''
    try:
        return  cv2.VideoCapture(str(path))
    except Exception:
        logger.warning(f'Failed to open cap for video at: {path}')
        return None

@pathify
@raise_on_path_not_exists
def get_cap_from_images_folder(folder, img_format="%1d.png"):
    """
        Creates a video capture from a folder of images.

        Argiments:
            folder: str, path to folder with images
            img_format: str. Format of the file names
    """
    # Create video capture
    cap = cv2.VideoCapture(str(folder / img_format))

    # Check all went well
    ret, frame = cap.read()
    if not ret:
        raise ValueError(
            "Something went wrong, can't read form folder: " + folder
        )
    else:
        cap_set_frame(cap, 0)
    return cap


# --------------------------------- trim clip -------------------------------- #
@open_video
def trim_clip(
    video, savepath, start_frame=0, end_frame=-1, fps=None,
):
    """trim_clip [take a videopath, open it and save a trimmed version between start and stop. Either 
    looking at a proportion of video (e.g. second half) or at start and stop frames]
    
    Arguments:
        videopath {[str]} -- [video to process]
        savepath {[str]} -- [where to save]
    
    Keyword Arguments:
        frame_mode {bool} -- [define start and stop time as frame numbers] (default: {False})
        start_frame {int} -- [video frame to stat at ] (default: {None})
        end_frame {int} -- [videoframe to stop at ] (default: {None})
        fps {[int]}(default, None) -- [specify the fps of the output]
    """

    # Open input file and extract params
    nframes, width, height, _fps, is_color = get_video_params(video)
    fps = fps or _fps

    # open writer
    writer = open_cvwriter(
        str(savepath),
        w=width,
        h=height,
        framerate=int(fps),
        format=".mp4",
        iscolor=is_color,
    )

    # Loop over frames and save the ones that matter
    cap_set_frame(video, start_frame)
    for framen in range(start_frame, end_frame):
        ret, frame = video.read()
        if not ret:
            logger.debug(
                f"FCUTILS: failed to read frame {framen} while trimming clip [{nframes} frames long]."
            )
            break
        else:
            if not is_color:
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            writer.write(frame)
    writer.release()



from scipy import stats
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit

from fcutils.plot.elements import plot_shaded_withline
from fcutils.maths.distributions import fit_kde


def plot_distribution(
    *args,
    dist_type="logistic",
    comulative=False,
    ax=None,
    shaded=False,
    shade_alpha=0.5,
    line_alpha=1,
    vertical=False,
    flip_y=False,
    x_range=None,
    plot_kwargs={},
    ax_kwargs={},
    fill_offset=0,
    y_scale=1,
    **kwargs,
):
    # Get the distribution
    if dist_type == "logistic":
        dist = stats.logistic(*args, **kwargs)
    elif dist_type == "gamma":
        dist = stats.gamma(*args, **kwargs)
    elif dist_type == "beta":
        dist = stats.beta(*args, **kwargs)
    elif (
        dist_type == "normal" or dist_type == "gaussian" or dist_type == "norm"
    ):
        dist = stats.norm(*args, **kwargs)
    elif dist_type == "exponential":
        dist = np.exp
    else:
        raise NotImplementedError

    # Get the probability density function or comulative density function
    if comulative:
        func = dist.cdf
    else:
        try:
            func = dist.pdf
            if not flip_y:
                x = np.linspace(dist.ppf(0.0001), dist.ppf(0.99999), 100)
            else:
                x = np.linspace(dist.ppf(0.0001), -dist.ppf(0.99999), 100)

        except:
            func = dist
            if not flip_y:
                x = np.linspace(x_range[0], x_range[1], 100)
            else:
                x = np.linspace(x_range[0], -x_range[1], 100)

    # Plot
    if ax is None:
        f, ax = plt.subplots()

    if not shaded:
        if not vertical:
            if not flip_y:
                ax.plot(x, func(x) * y_scale, **plot_kwargs)
            else:
                ax.plot(x, -func(x) * y_scale, **plot_kwargs)

        else:
            if not flip_y:
                ax.plot(func(x), x * y_scale, **plot_kwargs)
            else:
                ax.plot(func(x), -x * y_scale, **plot_kwargs)

    else:
        if not vertical:
            if not flip_y:
                ax.fill_between(
                    x,
                    fill_offset,
                    (func(x) + fill_offset) * y_scale,
                    alpha=shade_alpha,
                    **plot_kwargs,
                )
                ax.plot(
                    x,
                    (func(x) + fill_offset) * y_scale,
                    alpha=line_alpha,
                    **plot_kwargs,
                )
            else:
                ax.fill_between(
                    x,
                    fill_offset,
                    -func(x) + fill_offset * y_scale,
                    alpha=shade_alpha,
                    **plot_kwargs,
                )
                ax.plot(
                    x,
                    -(func(x) + fill_offset) * y_scale,
                    alpha=line_alpha,
                    **plot_kwargs,
                )
        else:
            if not flip_y:
                ax.fill_between(
                    (func(x)) * y_scale + fill_offset,
                    fill_offset,
                    x,
                    alpha=shade_alpha,
                    **plot_kwargs,
                )
                ax.plot(
                    (func(x)) * y_scale + fill_offset,
                    x,
                    alpha=line_alpha,
                    **plot_kwargs,
                )
            else:
                ax.fill_between(
                    (func(x)) * y_scale + fill_offset,
                    fill_offset,
                    -x,
                    alpha=shade_alpha,
                    **plot_kwargs,
                )
                ax.plot(
                    (func(x)) * y_scale + fill_offset,
                    -x,
                    alpha=line_alpha,
                    **plot_kwargs,
                )

    ax.set(**ax_kwargs)

    return ax


def plot_fitted_curve(
    func,
    xdata,
    ydata,
    ax,
    *args,
    xrange=None,
    print_fit=False,
    numpy_polyfit=False,
    fit_kwargs={},
    scatter_kwargs={},
    line_kwargs={},
):
    if numpy_polyfit and not isinstance(numpy_polyfit, int):
        raise ValueError("numpy_polyfit should be an integer")
    # set numpy_polifit to an integer to fit a numpy polinomial with degree numpy polyfit

    if xrange is not None:
        x = np.linspace(xrange[0], xrange[1], 100)
    else:
        x = np.linspace(np.min(xdata), np.max(xdata), 100)

    if not numpy_polyfit:  # ? scipy curve fit instead
        popt, pcov = curve_fit(func, xdata, ydata, *args, **fit_kwargs)
        if print_fit:
            print(popt)
        y = func(x, *popt)
        to_return = popt

    else:
        func = np.poly1d(np.polyfit(xdata, ydata, numpy_polyfit))
        y = func(x)
        to_return = func
    ax.scatter(xdata, ydata, **scatter_kwargs)
    ax.plot(x, y, **line_kwargs)

    return to_return


def plot_kde(
    ax=None,
    z=0,
    kde=None,
    data=None,
    invert=False,
    vertical=False,
    normto=None,
    fig_kwargs={},
    kde_kwargs={},
    **kwargs,
):
    """[Plots a KDE distribution. Plots first the shaded area and then the outline. 
       KDE can be oriented vertically, inverted, normalised...]
    
    Arguments:
        ax {[plt.axis]} -- [ax onto which to plot]
        kde {[np.ndarray]} -- [KDE fitted with statsmodels, optional. Either KDE or data must be passed]
        data {[np.ndarray]} -- [1d numpy array with data (Y values of distribution), optional. Either KDE or data must be passed]
        z {[type]} -- [value used to shift the curve, for a horizontal KDE z=0 means the curve is on the x axis. ]
    
    Keyword Arguments:
        invert {bool} -- [mirror the KDE plot relative to the X or Y axis, depending on ortentation] (default: {False})
        vertical {bool} -- [plot KDE vertically] (default: {False})
        normto {[float]} -- [normalise the KDE so that the peak of the distribution is at a certain value] (default: {None})
        label {[string]} -- [label for the legend] (default: {None})
    
    Returns:
        ax, kde
    """
    if ax is None:
        f, ax = plt.subplots(**fig_kwargs)

    if kde is None:
        if data is None:
            raise ValueError("either KDE or data must be passed")
        kde = fit_kde(data, **kde_kwargs)

    if vertical:
        x = kde.density
        y = kde.support
    else:
        x, y = kde.support, kde.density

    if normto is not None:
        if not vertical:
            y = y / np.max(y) * normto
        else:
            x = x / np.max(x) * normto

    if invert:
        y = z - y
    else:
        if not vertical:
            y = y + z
        else:
            x = x + z

    plot_shaded_withline(ax, x, y, z, **kwargs)

    return ax, kde



import cv2
from pathlib import Path


def open_video(func):
    """
        Decorator. If the first argument to a function is a path
        to a video it returns the cv2.VideoCapture object for it
    """

    def inner(*args, **kwargs):
        if isinstance(args[0], (str, Path)):
            args = list(args)
            args[0] = cv2.VideoCapture(str(args[0]))
            return func(*args, **kwargs)
        else:
            return func(*args, **kwargs)

    return inner



from fcutils.maths import array
import numpy as np


def test_find_nearest():
    arr = np.random.uniform(-100, 100, 100 * 100).reshape(100, 100)
    arr[40, 40] = 1000

    idx = array.find_nearest(arr, 1000)
    assert idx == 4040


def test_find_percentile():
    arr = np.random.uniform(-100, 100, 100 * 100).reshape(100, 100)
    arr[40, 40] = 1000

    array.percentile_range(arr)



import math
import numpy as np
from scipy import stats
import statsmodels.api as sm


def fit_kde(x, **kwargs):
    """ Fit a KDE using StatsModels. 
        kwargs is useful to pass stuff to the fit, e.g. the binwidth (bw)"""
    x = np.array(x).astype(np.float)
    kde = sm.nonparametric.KDEUnivariate(x)
    kde.fit(**kwargs)  # Estimate the densities
    return kde


def get_distribution(dist, *args, n_samples=10000):
    if dist == "uniform":
        return np.random.uniform(args[0], args[1], n_samples)
    elif dist == "normal":
        return np.random.normal(args[0], args[1], n_samples)
    elif dist == "beta":
        return np.random.beta(args[0], args[1], n_samples)
    elif dist == "gamma":
        return np.random.gamma(args[0], args[1], n_samples)


def get_parametric_distribution(
    dist, *args, x0=0.000001, x1=0.9999999, **kwargs
):
    if dist == "beta":
        dist = stats.beta(*args, **kwargs)
    elif dist == "normal" or dist == "norm" or dist == "gaussian":
        dist = stats.norm(*args, **kwargs)
    else:
        raise NotImplementedError

    support = np.linspace(dist.ppf(x0), dist.ppf(x1), 100)
    density = dist.pdf(support)
    return dist, support, density


# --------------------- get parameters for distirbutions --------------------- #
def beta_distribution_params(
    a=None, b=None, mu=None, sigma=None, omega=None, kappa=None
):
    """[converts parameters of beta into different formulations]
    
    Keyword Arguments:
        a {[type]} -- [a param] (default: {None})
        b {[type]} -- [b param] (default: {None})
        mu {[type]} -- [mean] (default: {None})
        sigma {[type]} -- [standard var] (default: {None})
        omega {[type]} -- [mode] (default: {None})
        kappa {[type]} -- [concentration] (default: {None})
    
    Raises:
        NotImplementedError: [description]
    """
    if kappa is not None and omega is not None:
        a = omega * (kappa - 2) + 1
        b = (1 - omega) * (kappa - 2) + 1
        return a, b
    elif a is not None and b is not None:
        mu = a / (a + b)
        omega = (a - 1) / (a + b - 2)
        kappa = a + b
        return mu, omega, kappa
    else:
        raise NotImplementedError


def gamma_distribution_params(
    mean=None, sd=None, mode=None, shape=None, rate=None
):
    if mean is not None and sd is not None:
        if mean < 0:
            raise NotImplementedError

        shape = mean ** 2 / sd ** 2
        rate = mean / sd ** 2
    elif mode is not None and sd is not None:
        if mode < 0:
            raise NotImplementedError

        rate = (mode + math.sqrt(mode ** 2 + 4 * (sd ** 2))) / (2 * (sd ** 2))
        shape = 1 + mode * rate
    elif shape is not None and rate is not None:
        mu = shape / rate
        sd = math.sqrt(shape) / rate
        return mu, sd
    return shape, rate



from datetime import datetime


def timestamp(just_time=False):
    if not just_time:
        return datetime.now().strftime("%Y-%m-%d  %H:%M")
    else:
        return datetime.now().strftime("%H:%M:%S")


if __name__ == "__main__":
    print(timestamp())



import numpy as np


def R(theta):
    """
        Returns the rotation matrix for rotating an object
        centered around the origin with a given angle

        Arguments:
            theta: angle in degrees

        Returns:
            R: 2x2 np.ndarray with rotation matrix
    """
    theta = np.radians(theta)
    return np.array(
        [[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]]
    )


def M(axis="x"):
    """
        Returns a matrix to mirror an object against a given axis

        Arguments:
            axis: str. 'x', 'y', 'origin' or 'xy'

        Returns:
            M: mirror matrix
    """
    if axis == "x":
        return np.array([[1, 0], [0, -1]])
    elif axis == "y":
        return np.array([[-1, 0], [0, 1]])
    elif axis == "origin":
        return np.array([[-1, 0], [0, -1]])
    elif axis == "xy":
        return np.array([[0, 1], [1, 0]])
    else:
        raise NotImplementedError(
            f"Could not recognize axis of mirroring: {axis}"
        )


def cart2pol(x, y):
    """
        Cartesian to polar coordinates

        angles in degrees
    """
    rho = np.hypot(x, y)
    phi = np.degrees(np.arctan2(y, x))
    return rho, phi


def pol2cart(rho, phi):
    """
        Polar to cartesian coordinates

        angles in degrees
    """
    x = rho * np.cos(np.radians(phi))
    y = rho * np.sin(np.radians(phi))
    return x, y



from setuptools import setup, find_namespace_packages

requirements = [
    "numpy",
    "opencv-python",
    "pandas",
    "matplotlib",
    "seaborn",
    "scipy",
    "pyyaml",
    "statsmodels",
    "loguru",
]

setup(
    name="fcutils",
    version="1.1.2.8",
    author_email="federicoclaudi@protonmail.com",
    description="bunch of utility functions",
    packages=find_namespace_packages(exclude=()),
    include_package_data=True,
    url="https://github.com/FedeClaudi/fcutils",
    author="Federico Claudi",
    zip_safe=False,
    install_requires=requirements,
)



from pathlib import Path


def sizeof_fmt(num, suffix="B"):
    """
        Returns a size in bytes to a human
        readable format
    """
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z", "Y"]:
        if abs(num) < 1024.0:
            return f"{num:.2f} {unit}{suffix}"
        num /= 1024.0


def raise_on_path_not_exists(func):
    """
        Decorator that raises an error when the
        path passed to a function does not point towards an actual file
    """

    def inner(*args, **kwargs):
        args = list(args)
        path = args[0]
        if not path.exists():
            raise FileNotFoundError(f"File or folder: {path} does not exist")
        else:
            return func(*args, **kwargs)

    return inner


def pathify(func):
    """ 
        Decorator that makes sure that the first arugment to a function
        is a Path object and not a path as string
    """

    def inner(*args, **kwargs):
        args = list(args)
        args[0] = Path(args[0])
        return func(*args, **kwargs)

    return inner


def return_list_smart(func):
    """ 
        Decorator. If a function is trying to return a list
        of length 1 it returns the list element instead
    """

    def inner(*args, **kwargs):
        out = func(*args, **kwargs)
        if not out:
            return None

        if len(out) == 1:
            return out[0]
        else:
            return out

    return inner
